"""Generated wrapper for MultiAssetProxy Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for MultiAssetProxy below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        MultiAssetProxyValidator,
    )
except ImportError:

    class MultiAssetProxyValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass


class AssetProxiesMethod(ContractMethod):
    """Various interfaces to the assetProxies method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: Union[bytes, str]):
        """Validate the inputs to the assetProxies method."""
        self.validator.assert_valid(
            method_name="assetProxies",
            parameter_name="index_0",
            argument_value=index_0,
        )
        return index_0

    def call(
        self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).call(tx_params.as_dict())

    def send_transaction(
        self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).transact(tx_params.as_dict())

    def estimate_gas(
        self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).estimateGas(tx_params.as_dict())


class AddAuthorizedAddressMethod(ContractMethod):
    """Various interfaces to the addAuthorizedAddress method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, target: str):
        """Validate the inputs to the addAuthorizedAddress method."""
        self.validator.assert_valid(
            method_name="addAuthorizedAddress",
            parameter_name="target",
            argument_value=target,
        )
        target = self.validate_and_checksum_address(target)
        return target

    def call(
        self, target: str, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Authorizes an address.

        :param target: Address to authorize.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (target) = self.validate_and_normalize_inputs(target)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target).call(tx_params.as_dict())

    def send_transaction(
        self, target: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Authorizes an address.

        :param target: Address to authorize.
        :param tx_params: transaction parameters
        """
        (target) = self.validate_and_normalize_inputs(target)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target).transact(tx_params.as_dict())

    def estimate_gas(
        self, target: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (target) = self.validate_and_normalize_inputs(target)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target).estimateGas(tx_params.as_dict())


class AuthoritiesMethod(ContractMethod):
    """Various interfaces to the authorities method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: int):
        """Validate the inputs to the authorities method."""
        self.validator.assert_valid(
            method_name="authorities",
            parameter_name="index_0",
            argument_value=index_0,
        )
        # safeguard against fractional inputs
        index_0 = int(index_0)
        return index_0

    def call(self, index_0: int, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).call(tx_params.as_dict())

    def send_transaction(
        self, index_0: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).transact(tx_params.as_dict())

    def estimate_gas(
        self, index_0: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).estimateGas(tx_params.as_dict())


class GetAssetProxyMethod(ContractMethod):
    """Various interfaces to the getAssetProxy method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, asset_proxy_id: Union[bytes, str]):
        """Validate the inputs to the getAssetProxy method."""
        self.validator.assert_valid(
            method_name="getAssetProxy",
            parameter_name="assetProxyId",
            argument_value=asset_proxy_id,
        )
        return asset_proxy_id

    def call(
        self,
        asset_proxy_id: Union[bytes, str],
        tx_params: Optional[TxParams] = None,
    ) -> str:
        """Execute underlying contract method via eth_call.

        Gets an asset proxy.

        :param assetProxyId: Id of the asset proxy.
        :param tx_params: transaction parameters
        :returns: The asset proxy registered to assetProxyId. Returns 0x0 if no
            proxy is registered.
        """
        (asset_proxy_id) = self.validate_and_normalize_inputs(asset_proxy_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_proxy_id).call(tx_params.as_dict())

    def send_transaction(
        self,
        asset_proxy_id: Union[bytes, str],
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Gets an asset proxy.

        :param assetProxyId: Id of the asset proxy.
        :param tx_params: transaction parameters
        :returns: The asset proxy registered to assetProxyId. Returns 0x0 if no
            proxy is registered.
        """
        (asset_proxy_id) = self.validate_and_normalize_inputs(asset_proxy_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_proxy_id).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        asset_proxy_id: Union[bytes, str],
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (asset_proxy_id) = self.validate_and_normalize_inputs(asset_proxy_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_proxy_id).estimateGas(
            tx_params.as_dict()
        )


class RemoveAuthorizedAddressMethod(ContractMethod):
    """Various interfaces to the removeAuthorizedAddress method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, target: str):
        """Validate the inputs to the removeAuthorizedAddress method."""
        self.validator.assert_valid(
            method_name="removeAuthorizedAddress",
            parameter_name="target",
            argument_value=target,
        )
        target = self.validate_and_checksum_address(target)
        return target

    def call(
        self, target: str, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Removes authorizion of an address.

        :param target: Address to remove authorization from.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (target) = self.validate_and_normalize_inputs(target)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target).call(tx_params.as_dict())

    def send_transaction(
        self, target: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Removes authorizion of an address.

        :param target: Address to remove authorization from.
        :param tx_params: transaction parameters
        """
        (target) = self.validate_and_normalize_inputs(target)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target).transact(tx_params.as_dict())

    def estimate_gas(
        self, target: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (target) = self.validate_and_normalize_inputs(target)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target).estimateGas(tx_params.as_dict())


class OwnerMethod(ContractMethod):
    """Various interfaces to the owner method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().estimateGas(tx_params.as_dict())


class RemoveAuthorizedAddressAtIndexMethod(ContractMethod):
    """Various interfaces to the removeAuthorizedAddressAtIndex method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, target: str, index: int):
        """Validate the inputs to the removeAuthorizedAddressAtIndex method."""
        self.validator.assert_valid(
            method_name="removeAuthorizedAddressAtIndex",
            parameter_name="target",
            argument_value=target,
        )
        target = self.validate_and_checksum_address(target)
        self.validator.assert_valid(
            method_name="removeAuthorizedAddressAtIndex",
            parameter_name="index",
            argument_value=index,
        )
        # safeguard against fractional inputs
        index = int(index)
        return (target, index)

    def call(
        self, target: str, index: int, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Removes authorizion of an address.

        :param index: Index of target in authorities array.
        :param target: Address to remove authorization from.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (target, index) = self.validate_and_normalize_inputs(target, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target, index).call(tx_params.as_dict())

    def send_transaction(
        self, target: str, index: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Removes authorizion of an address.

        :param index: Index of target in authorities array.
        :param target: Address to remove authorization from.
        :param tx_params: transaction parameters
        """
        (target, index) = self.validate_and_normalize_inputs(target, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target, index).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, target: str, index: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (target, index) = self.validate_and_normalize_inputs(target, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(target, index).estimateGas(
            tx_params.as_dict()
        )


class GetProxyIdMethod(ContractMethod):
    """Various interfaces to the getProxyId method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        Gets the proxy id associated with the proxy address.

        :param tx_params: transaction parameters
        :returns: Proxy id.
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Gets the proxy id associated with the proxy address.

        :param tx_params: transaction parameters
        :returns: Proxy id.
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().estimateGas(tx_params.as_dict())


class AuthorizedMethod(ContractMethod):
    """Various interfaces to the authorized method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: str):
        """Validate the inputs to the authorized method."""
        self.validator.assert_valid(
            method_name="authorized",
            parameter_name="index_0",
            argument_value=index_0,
        )
        index_0 = self.validate_and_checksum_address(index_0)
        return index_0

    def call(self, index_0: str, tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).call(tx_params.as_dict())

    def send_transaction(
        self, index_0: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).transact(tx_params.as_dict())

    def estimate_gas(
        self, index_0: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(index_0).estimateGas(tx_params.as_dict())


class RegisterAssetProxyMethod(ContractMethod):
    """Various interfaces to the registerAssetProxy method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, asset_proxy: str):
        """Validate the inputs to the registerAssetProxy method."""
        self.validator.assert_valid(
            method_name="registerAssetProxy",
            parameter_name="assetProxy",
            argument_value=asset_proxy,
        )
        asset_proxy = self.validate_and_checksum_address(asset_proxy)
        return asset_proxy

    def call(
        self, asset_proxy: str, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Registers an asset proxy to its asset proxy id. Once an asset proxy is
        registered, it cannot be unregistered.

        :param assetProxy: Address of new asset proxy to register.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (asset_proxy) = self.validate_and_normalize_inputs(asset_proxy)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_proxy).call(tx_params.as_dict())

    def send_transaction(
        self, asset_proxy: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Registers an asset proxy to its asset proxy id. Once an asset proxy is
        registered, it cannot be unregistered.

        :param assetProxy: Address of new asset proxy to register.
        :param tx_params: transaction parameters
        """
        (asset_proxy) = self.validate_and_normalize_inputs(asset_proxy)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_proxy).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, asset_proxy: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (asset_proxy) = self.validate_and_normalize_inputs(asset_proxy)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_proxy).estimateGas(
            tx_params.as_dict()
        )


class GetAuthorizedAddressesMethod(ContractMethod):
    """Various interfaces to the getAuthorizedAddresses method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> List[str]:
        """Execute underlying contract method via eth_call.

        Gets all authorized addresses.

        :param tx_params: transaction parameters
        :returns: Array of authorized addresses.
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Gets all authorized addresses.

        :param tx_params: transaction parameters
        :returns: Array of authorized addresses.
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().estimateGas(tx_params.as_dict())


class TransferOwnershipMethod(ContractMethod):
    """Various interfaces to the transferOwnership method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, new_owner: str):
        """Validate the inputs to the transferOwnership method."""
        self.validator.assert_valid(
            method_name="transferOwnership",
            parameter_name="newOwner",
            argument_value=new_owner,
        )
        new_owner = self.validate_and_checksum_address(new_owner)
        return new_owner

    def call(
        self, new_owner: str, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(new_owner).call(tx_params.as_dict())

    def send_transaction(
        self, new_owner: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(new_owner).transact(tx_params.as_dict())

    def estimate_gas(
        self, new_owner: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(new_owner).estimateGas(
            tx_params.as_dict()
        )


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class MultiAssetProxy:
    """Wrapper class for MultiAssetProxy Solidity contract."""

    asset_proxies: AssetProxiesMethod
    """Constructor-initialized instance of
    :class:`AssetProxiesMethod`.
    """

    add_authorized_address: AddAuthorizedAddressMethod
    """Constructor-initialized instance of
    :class:`AddAuthorizedAddressMethod`.
    """

    authorities: AuthoritiesMethod
    """Constructor-initialized instance of
    :class:`AuthoritiesMethod`.
    """

    get_asset_proxy: GetAssetProxyMethod
    """Constructor-initialized instance of
    :class:`GetAssetProxyMethod`.
    """

    remove_authorized_address: RemoveAuthorizedAddressMethod
    """Constructor-initialized instance of
    :class:`RemoveAuthorizedAddressMethod`.
    """

    owner: OwnerMethod
    """Constructor-initialized instance of
    :class:`OwnerMethod`.
    """

    remove_authorized_address_at_index: RemoveAuthorizedAddressAtIndexMethod
    """Constructor-initialized instance of
    :class:`RemoveAuthorizedAddressAtIndexMethod`.
    """

    get_proxy_id: GetProxyIdMethod
    """Constructor-initialized instance of
    :class:`GetProxyIdMethod`.
    """

    authorized: AuthorizedMethod
    """Constructor-initialized instance of
    :class:`AuthorizedMethod`.
    """

    register_asset_proxy: RegisterAssetProxyMethod
    """Constructor-initialized instance of
    :class:`RegisterAssetProxyMethod`.
    """

    get_authorized_addresses: GetAuthorizedAddressesMethod
    """Constructor-initialized instance of
    :class:`GetAuthorizedAddressesMethod`.
    """

    transfer_ownership: TransferOwnershipMethod
    """Constructor-initialized instance of
    :class:`TransferOwnershipMethod`.
    """

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: MultiAssetProxyValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = MultiAssetProxyValidator(
                web3_or_provider, contract_address
            )

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        if web3 is None:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                        middleware["function"], layer=middleware["layer"]
                    )
            except ValueError as value_error:
                if value_error.args == (
                    "You can't add the same un-named instance twice",
                ):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(
            address=to_checksum_address(contract_address),
            abi=MultiAssetProxy.abi(),
        ).functions

        self.asset_proxies = AssetProxiesMethod(
            web3_or_provider,
            contract_address,
            functions.assetProxies,
            validator,
        )

        self.add_authorized_address = AddAuthorizedAddressMethod(
            web3_or_provider,
            contract_address,
            functions.addAuthorizedAddress,
            validator,
        )

        self.authorities = AuthoritiesMethod(
            web3_or_provider,
            contract_address,
            functions.authorities,
            validator,
        )

        self.get_asset_proxy = GetAssetProxyMethod(
            web3_or_provider,
            contract_address,
            functions.getAssetProxy,
            validator,
        )

        self.remove_authorized_address = RemoveAuthorizedAddressMethod(
            web3_or_provider,
            contract_address,
            functions.removeAuthorizedAddress,
            validator,
        )

        self.owner = OwnerMethod(
            web3_or_provider, contract_address, functions.owner, validator
        )

        self.remove_authorized_address_at_index = RemoveAuthorizedAddressAtIndexMethod(
            web3_or_provider,
            contract_address,
            functions.removeAuthorizedAddressAtIndex,
            validator,
        )

        self.get_proxy_id = GetProxyIdMethod(
            web3_or_provider, contract_address, functions.getProxyId, validator
        )

        self.authorized = AuthorizedMethod(
            web3_or_provider, contract_address, functions.authorized, validator
        )

        self.register_asset_proxy = RegisterAssetProxyMethod(
            web3_or_provider,
            contract_address,
            functions.registerAssetProxy,
            validator,
        )

        self.get_authorized_addresses = GetAuthorizedAddressesMethod(
            web3_or_provider,
            contract_address,
            functions.getAuthorizedAddresses,
            validator,
        )

        self.transfer_ownership = TransferOwnershipMethod(
            web3_or_provider,
            contract_address,
            functions.transferOwnership,
            validator,
        )

    def get_authorized_address_added_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for AuthorizedAddressAdded event.

        :param tx_hash: hash of transaction emitting AuthorizedAddressAdded
            event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=MultiAssetProxy.abi(),
            )
            .events.AuthorizedAddressAdded()
            .processReceipt(tx_receipt)
        )

    def get_authorized_address_removed_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for AuthorizedAddressRemoved event.

        :param tx_hash: hash of transaction emitting AuthorizedAddressRemoved
            event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=MultiAssetProxy.abi(),
            )
            .events.AuthorizedAddressRemoved()
            .processReceipt(tx_receipt)
        )

    def get_asset_proxy_registered_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for AssetProxyRegistered event.

        :param tx_hash: hash of transaction emitting AssetProxyRegistered event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=MultiAssetProxy.abi(),
            )
            .events.AssetProxyRegistered()
            .processReceipt(tx_receipt)
        )

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"constant":true,"inputs":[{"internalType":"bytes4","name":"index_0","type":"bytes4"}],"name":"assetProxies","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"target","type":"address"}],"name":"addAuthorizedAddress","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"internalType":"uint256","name":"index_0","type":"uint256"}],"name":"authorities","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"internalType":"bytes4","name":"assetProxyId","type":"bytes4"}],"name":"getAssetProxy","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"target","type":"address"}],"name":"removeAuthorizedAddress","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"target","type":"address"},{"internalType":"uint256","name":"index","type":"uint256"}],"name":"removeAuthorizedAddressAtIndex","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"getProxyId","outputs":[{"internalType":"bytes4","name":"","type":"bytes4"}],"payable":false,"stateMutability":"pure","type":"function"},{"constant":true,"inputs":[{"internalType":"address","name":"index_0","type":"address"}],"name":"authorized","outputs":[{"internalType":"bool","name":"","type":"bool"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"assetProxy","type":"address"}],"name":"registerAssetProxy","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"getAuthorizedAddresses","outputs":[{"internalType":"address[]","name":"","type":"address[]"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"newOwner","type":"address"}],"name":"transferOwnership","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"payable":false,"stateMutability":"nonpayable","type":"fallback"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"target","type":"address"},{"indexed":true,"internalType":"address","name":"caller","type":"address"}],"name":"AuthorizedAddressAdded","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"target","type":"address"},{"indexed":true,"internalType":"address","name":"caller","type":"address"}],"name":"AuthorizedAddressRemoved","type":"event"},{"anonymous":false,"inputs":[{"indexed":false,"internalType":"bytes4","name":"id","type":"bytes4"},{"indexed":false,"internalType":"address","name":"assetProxy","type":"address"}],"name":"AssetProxyRegistered","type":"event"}]'  # noqa: E501 (line-too-long)
        )


# pylint: disable=too-many-lines



"""Exchange-specific exception classes."""

from enum import auto, Enum

from zero_ex.contract_wrappers.exceptions import RichRevert

# pylint: disable=missing-docstring


class AssetProxyDispatchErrorCodes(Enum):  # noqa: D101 (missing docstring)
    INVALID_ASSET_DATA_LENGTH = 0
    UNKNOWN_ASSET_PROXY = auto()


class BatchMatchOrdersErrorCodes(Enum):  # noqa: D101 (missing docstring)
    ZERO_LEFT_ORDERS = 0
    ZERO_RIGHT_ORDERS = auto()
    INVALID_LENGTH_LEFT_SIGNATURES = auto()
    INVALID_LENGTH_RIGHT_SIGNATURES = auto()


class ExchangeContextErrorCodes(Enum):  # noqa: D101 (missing docstring)
    INVALID_MAKER = 0
    INVALID_TAKER = auto()
    INVALID_SENDER = auto()


class FillErrorCodes(Enum):  # noqa: D101 (missing docstring)
    INVALID_TAKER_AMOUNT = 0
    TAKER_OVERPAY = auto()
    OVERFILL = auto()
    INVALID_FILL_PRICE = auto()


class SignatureErrorCodes(Enum):  # noqa: D101 (missing docstring)
    BAD_ORDER_SIGNATURE = 0
    BAD_TRANSACTION_SIGNATURE = auto()
    INVALID_LENGTH = auto()
    UNSUPPORTED = auto()
    ILLEGAL = auto()
    INAPPROPRIATE_SIGNATURE_TYPE = auto()
    INVALID_SIGNER = auto()


class TransactionErrorCodes(Enum):  # noqa: D101 (missing docstring)
    ALREADY_EXECUTED = 0
    EXPIRED = auto()


class IncompleteFillErrorCode(Enum):  # noqa: D101 (missing docstring)
    INCOMPLETE_MARKET_BUY_ORDERS = 0
    INCOMPLETE_MARKET_SELL_ORDERS = auto()
    INCOMPLETE_FILL_ORDER = auto()


class SignatureError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "SignatureError(uint8,bytes32,address,bytes)",
            ["errorCode", "hash", "signerAddress", "signature"],
            return_data,
        )

    errorCode: SignatureErrorCodes
    hash: bytes
    signerAddress: str
    signature: bytes

    selector = "0x7e5a2318"


class SignatureValidatorNotApprovedError(
    RichRevert
):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "SignatureValidatorNotApprovedError(address,address)",
            ["signerAddress", "validatorAddress"],
            return_data,
        )

    signerAddress: str
    validatorAddress: str

    selector = "0xa15c0d06"


class EIP1271SignatureError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "EIP1271SignatureError(address,bytes,bytes,bytes)",
            ["verifyingContractAddress", "data", "signature", "errorData"],
            return_data,
        )

    verifyingContractAddress: str
    data: bytes
    signature: bytes
    errorData: bytes

    selector = "0x5bd0428d"


class SignatureWalletError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "SignatureWalletError(bytes32,address,bytes,bytes)",
            ["hash", "walletAddress", "signature", "errorData"],
            return_data,
        )

    hash: bytes
    walletAddress: str
    signature: bytes
    errorData: bytes

    selector = "0x1b8388f7"


class OrderStatusError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "OrderStatusError(bytes32,uint8)",
            ["orderHash", "orderStatus"],
            return_data,
        )

    orderHash: bytes
    orderStatus: int

    selector = "0xfdb6ca8d"


class ExchangeInvalidContextError(
    RichRevert
):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "ExchangeInvalidContextError(uint8,bytes32,address)",
            ["errorCode", "orderHash", "contextAddress"],
            return_data,
        )

    errorCode: ExchangeContextErrorCodes
    orderHash: bytes
    contextAddress: str

    selector = "0xe53c76c8"


class FillError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "FillError(uint8,bytes32)", ["errorCode", "orderHash"], return_data
        )

    errorCode: FillErrorCodes
    orderHash: bytes

    selector = "0xe94a7ed0"


class OrderEpochError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "OrderEpochError(address,address,uint256)",
            ["makerAddress", "orderSenderAddress", "currentEpoch"],
            return_data,
        )

    makerAddress: str
    orderSenderAddress: str
    currentEpoch: int

    selector = "0x4ad31275"


class AssetProxyExistsError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "AssetProxyExistsError(bytes4,address)",
            ["assetProxyId", "assetProxyAddress"],
            return_data,
        )

    assetProxyId: bytes
    assetProxyAddress: str

    selector = "0x11c7b720"


class AssetProxyDispatchError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "AssetProxyDispatchError(uint8,bytes32,bytes)",
            ["errorCode", "orderHash", "assetData"],
            return_data,
        )

    errorCode: AssetProxyDispatchErrorCodes
    orderHash: bytes
    assetData: bytes

    selector = "0x488219a6"


class AssetProxyTransferError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "AssetProxyTransferError(bytes32,bytes,bytes)",
            ["orderHash", "assetData", "errorData"],
            return_data,
        )

    orderHash: bytes
    assetData: bytes
    errorData: bytes

    selector = "0x4678472b"


class NegativeSpreadError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "NegativeSpreadError(bytes32,bytes32)",
            ["leftOrderHash", "rightOrderHash"],
            return_data,
        )

    leftOrderHash: bytes
    rightOrderHash: bytes

    selector = "0xb6555d6f"


class TransactionError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "TransactionError(uint8,bytes32)",
            ["errorCode", "transactionHash"],
            return_data,
        )

    errorCode: TransactionErrorCodes
    transactionHash: bytes

    selector = "0xf5985184"


class TransactionExecutionError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "TransactionExecutionError(bytes32,bytes)",
            ["transactionHash", "errorData"],
            return_data,
        )

    transactionHash: bytes
    errorData: bytes

    selector = "0x20d11f61"


class TransactionGasPriceError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "TransactionGasPriceError(bytes32,uint256,uint256)",
            ["transactionHash", "actualGasPrice", "requiredGasPrice"],
            return_data,
        )

    transactionHash: bytes
    actualGasPrice: int
    requiredGasPrice: int

    selector = "0xa26dac09"


class TransactionInvalidContextError(
    RichRevert
):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "TransactionInvalidContextError(bytes32,address)",
            ["transactionHash", "currentContextAddress"],
            return_data,
        )

    transactionHash: bytes
    currentContextAddress: str

    selector = "0xdec4aedf"


class IncompleteFillError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "IncompleteFillError(uint8,uint256,uint256)",
            ["errorCode", "expectedAssetAmount", "actualAssetAmount"],
            return_data,
        )

    errorCode: IncompleteFillErrorCode
    expectedAssetAmount: int
    actualAssetAmount: int

    selector = "0x18e4b141"


class BatchMatchOrdersError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "BatchMatchOrdersError(uint8)", ["errorCode"], return_data
        )

    errorCode: BatchMatchOrdersErrorCodes

    selector = "0xd4092f4f"


class PayProtocolFeeError(RichRevert):  # noqa: D101 (missing docstring)
    def __init__(self, return_data):  # noqa: D107 (missing docstring)
        super().__init__(
            "PayProtocolFeeError(bytes32,uint256,address,address,bytes)",
            [
                "orderHash",
                "protocolFee",
                "makerAddress",
                "takerAddress",
                "errorData",
            ],
            return_data,
        )

    orderHash: bytes
    protocolFee: int
    makerAddress: str
    takerAddress: str
    errorData: bytes

    selector = "0x87cb1e75"



"""Web3.py-compatible middleware to be injected upon contract instantiation."""

from zero_ex.contract_wrappers.exceptions import (
    exception_class_from_rich_revert_selector,
    NoExceptionForSelector,
)

from . import exceptions


def rich_revert_handler(make_request, _):
    """Return a middlware to raise exceptions for rich revert return data."""
    # noqa: D202 (No blank lines allowed after function docstring
    def middleware(method, params):
        response = make_request(method, params)
        try:
            raise exception_class_from_rich_revert_selector(
                response["result"][0:10], exceptions
            )(response["result"])
        except NoExceptionForSelector:
            # response prefix didn't indicate a known error
            pass
        except TypeError:
            # eg "unhashable type: 'slice'". if response["result"] isn't
            # sliceable (eg if it's a dict), then it definitely isn't a rich
            # revert.
            pass
        except KeyError:
            # response doesn't have a "result" key
            pass
        return response

    return middleware


MIDDLEWARE = [{"layer": 0, "function": rich_revert_handler}]



"""Validate inputs to the Exchange contract."""

from typing import Any, Union

from web3 import Web3
from web3.providers.base import BaseProvider

from zero_ex import json_schemas
from zero_ex.contract_wrappers.order_conversions import order_to_jsdict

from ..bases import Validator


class ExchangeValidator(Validator):
    """Validate inputs to Exchange methods."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
    ):
        """Initialize the class."""
        super().__init__(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        if web3 is None:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        self.contract_address = contract_address
        self.chain_id = web3.eth.chainId

    def assert_valid(
        self, method_name: str, parameter_name: str, argument_value: Any
    ) -> None:
        """Raise an exception if method input is not valid.

        :param method_name: Name of the method whose input is to be validated.
        :param parameter_name: Name of the parameter whose input is to be
            validated.
        :param argument_value: Value of argument to parameter to be validated.
        """
        if parameter_name == "order":
            json_schemas.assert_valid(
                order_to_jsdict(
                    argument_value, self.chain_id, self.contract_address
                ),
                "/orderSchema",
            )

        if parameter_name == "orders":
            for order in argument_value:
                json_schemas.assert_valid(
                    order_to_jsdict(
                        order, self.chain_id, self.contract_address
                    ),
                    "/orderSchema",
                )



"""Conveniences for handling types representing Exchange Solidity structs.

The `TypedDict`:code: classes in the .exchange module represent tuples
encountered in the Exchange contract's ABI.  However, they have weird names,
containing hashes of the tuple's field names, because the name of a Solidity
`struct`:code: isn't conveyed through the ABI.  This module provides type
aliases with human-friendly names.

Converting between the JSON wire format and the types accepted by Web3.py (eg
`bytes` vs `str`) can be onerous.  This module provides conveniences for
converting Exchange structs between JSON and Python objects.
"""

from enum import auto, Enum

from . import (
    Tuple0x735c43e3,
    Tuple0x6ca34a6f,
    Tuple0x4c5ca29b,
    Tuple0xdabc15fe,
    Tuple0xb1e4a1ae,
)


# Would rather not repeat ourselves below, but base classes are mentioned in
# the class docstrings because of a bug in sphinx rendering.  Using the `..
# autoclass` directive, with the `:show-inheritance:` role, results in docs
# being rendered with just "Bases: dict", and no mention of the direct ancestor
# of each of these classes.


class FillResults(Tuple0x735c43e3):
    """The `FillResults`:code: Solidity struct.

    Also known as
    `zero_ex.contract_wrappers.exchange.Tuple0x735c43e3`:py:class:.
    """


class Order(Tuple0x6ca34a6f):
    """The `Order`:code: Solidity struct.

    Also known as
    `zero_ex.contract_wrappers.exchange.Tuple0x6ca34a6f`:py:class:.
    """


class MatchedFillResults(Tuple0x4c5ca29b):
    """The `MatchedFillResults`:code: Solidity struct.

    Also known as
    `zero_ex.contract_wrappers.exchange.Tuple0x4c5ca29b`:py:class:.
    """


class ZeroExTransaction(Tuple0xdabc15fe):
    """The `ZeroExTransaction`:code: Solidity struct.

    Also known as
    `zero_ex.contract_wrappers.exchange.Tuple0xdabc15fe`:py:class:.
    """


class OrderInfo(Tuple0xb1e4a1ae):
    """The `OrderInfo`:code: Solidity struct.

    Also known as
    `zero_ex.contract_wrappers.exchange.Tuple0xb1e4a1ae`:py:class:.
    """


class OrderStatus(Enum):  # noqa: D101 # pylint: disable=missing-docstring
    INVALID = 0
    INVALID_MAKER_ASSET_AMOUNT = auto()
    INVALID_TAKER_ASSET_AMOUNT = auto()
    FILLABLE = auto()
    EXPIRED = auto()
    FULLY_FILLED = auto()
    CANCELLED = auto()



"""Test 0x Exchnage wrapper."""

import random

import pytest
from eth_utils import remove_0x_prefix

from zero_ex.contract_addresses import NETWORK_TO_ADDRESSES, NetworkId
from zero_ex.contract_wrappers import TxParams
from zero_ex.contract_wrappers.exchange import Exchange
from zero_ex.contract_wrappers.exchange.types import Order
from zero_ex.json_schemas import assert_valid
from zero_ex.order_utils import (
    asset_data_utils,
    generate_order_hash_hex,
    sign_hash,
)


@pytest.fixture(scope="module")
def exchange_wrapper(ganache_provider):
    """Get an Exchange wrapper instance."""
    return Exchange(
        web3_or_provider=ganache_provider,
        contract_address=NETWORK_TO_ADDRESSES[NetworkId.GANACHE].exchange,
    )


def create_test_order(
    maker_address,
    maker_asset_amount,
    maker_asset_data,
    taker_asset_amount,
    taker_asset_data,
):
    """Create a test order."""
    order = Order(
        makerAddress=maker_address,
        takerAddress="0x0000000000000000000000000000000000000000",
        feeRecipientAddress="0x0000000000000000000000000000000000000000",
        senderAddress="0x0000000000000000000000000000000000000000",
        makerAssetAmount=maker_asset_amount,
        takerAssetAmount=taker_asset_amount,
        makerFee=0,
        takerFee=0,
        expirationTimeSeconds=100000000000000,
        salt=random.randint(1, 1000000000),
        makerAssetData=maker_asset_data,
        takerAssetData=taker_asset_data,
        makerFeeAssetData=asset_data_utils.encode_erc20("0x" + "00" * 20),
        takerFeeAssetData=asset_data_utils.encode_erc20("0x" + "00" * 20),
    )
    return order


def assert_fill_log(fill_log, maker, taker, order, order_hash):
    """assert that the fill log matches the order details"""
    assert fill_log.makerAddress == maker
    assert fill_log.takerAddress == taker
    assert fill_log.feeRecipientAddress == order["feeRecipientAddress"]
    assert fill_log.senderAddress == taker
    assert fill_log.orderHash == bytes.fromhex(remove_0x_prefix(order_hash))
    assert fill_log.makerAssetFilledAmount == order["makerAssetAmount"]
    assert fill_log.takerAssetFilledAmount == order["takerAssetAmount"]
    assert fill_log.makerFeePaid == order["makerFee"]
    assert fill_log.takerFeePaid == order["takerFee"]
    assert fill_log.makerAssetData == order["makerAssetData"]
    assert fill_log.takerAssetData == order["takerAssetData"]


def test_exchange_wrapper__fill_order(
    accounts,
    exchange_wrapper,  # pylint: disable=redefined-outer-name
    ganache_provider,
    weth_asset_data,
    zrx_asset_data,
):
    """Test filling an order."""
    taker = accounts[0]
    maker = accounts[1]
    exchange_address = exchange_wrapper.contract_address
    order = create_test_order(maker, 1, weth_asset_data, 1, zrx_asset_data)
    order_hash = generate_order_hash_hex(
        order=order, exchange_address=exchange_address, chain_id=1337
    )
    order_signature = sign_hash(ganache_provider, maker, order_hash)

    fill_results = exchange_wrapper.fill_order.call(
        order=order,
        taker_asset_fill_amount=order["takerAssetAmount"],
        signature=order_signature,
        tx_params=TxParams(from_=taker),
    )
    assert fill_results[0] == 1
    assert fill_results[1] == 1
    assert fill_results[2] == 0
    assert fill_results[3] == 0
    assert fill_results[4] == 0

    tx_hash = exchange_wrapper.fill_order.send_transaction(
        order=order,
        taker_asset_fill_amount=order["takerAssetAmount"],
        signature=order_signature,
        tx_params=TxParams(from_=taker),
    )
    assert_valid(tx_hash.hex(), "/hexSchema")

    fill_event = exchange_wrapper.get_fill_event(tx_hash)
    assert_fill_log(fill_event[0].args, maker, taker, order, order_hash)


# pylint: disable=too-many-locals
def test_exchange_wrapper__batch_fill_orders(
    accounts,
    exchange_wrapper,  # pylint: disable=redefined-outer-name
    ganache_provider,
    weth_asset_data,
):
    """Test filling a batch of orders."""
    taker = accounts[0]
    maker = accounts[1]
    exchange_address = exchange_wrapper.contract_address
    orders = []
    order_1 = create_test_order(maker, 1, weth_asset_data, 1, weth_asset_data)
    order_2 = create_test_order(maker, 1, weth_asset_data, 1, weth_asset_data)
    orders.append(order_1)
    orders.append(order_2)
    order_hashes = [
        generate_order_hash_hex(
            order=order, exchange_address=exchange_address, chain_id=1337
        )
        for order in orders
    ]
    order_signatures = [
        sign_hash(ganache_provider, maker, order_hash)
        for order_hash in order_hashes
    ]
    taker_amounts = [order["takerAssetAmount"] for order in orders]
    tx_hash = exchange_wrapper.batch_fill_orders.send_transaction(
        orders=orders,
        taker_asset_fill_amounts=taker_amounts,
        signatures=order_signatures,
        tx_params=TxParams(from_=taker),
    )
    assert_valid(tx_hash.hex(), "/hexSchema")

    fill_events = exchange_wrapper.get_fill_event(tx_hash)
    for index, order in enumerate(orders):
        assert_fill_log(
            fill_events[index].args, maker, taker, order, order_hashes[index]
        )


def test_two_instantiations_with_web3_objects(web3_instance):
    """Test that instantiating two Exchange objects doesn't raise.

    When instantiating an Exchange object with a web3 client (rather than a
    provider) there was a bug encountered where web3.py was giving an error
    when trying to install the rich-revert-handling middleware on the web3
    client, an error saying "can't install this same middleware instance
    again."  Test that that bug isn't occurring.
    """
    exchange = Exchange(  # pylint: disable=unused-variable
        web3_instance, NETWORK_TO_ADDRESSES[NetworkId.GANACHE].exchange
    )
    exchange2 = Exchange(  # pylint: disable=unused-variable
        web3_instance, NETWORK_TO_ADDRESSES[NetworkId.GANACHE].exchange
    )



"""Exception classes common to all wrappers."""

from inspect import isclass
from typing import List

from eth_abi import decode_abi


class RichRevert(Exception):
    """Raised when a contract method returns a rich revert error."""

    def __init__(
        self, abi_signature: str, param_names: List[str], return_data: str
    ):
        """Populate instance variables with decoded return data values."""
        arg_start_index = abi_signature.index("(") + 1
        arg_end_index = abi_signature.index(")")
        arguments = decode_abi(
            abi_signature[arg_start_index:arg_end_index].split(","),
            bytes.fromhex(return_data[10:]),
        )
        for (param_name, argument) in zip(param_names, arguments):
            setattr(self, param_name, argument)
        super().__init__(vars(self))


class NoExceptionForSelector(Exception):
    """Indicates that no exception could be found for the given selector."""


def exception_class_from_rich_revert_selector(
    selector: str, exceptions_module
) -> RichRevert:
    """Return the appropriate exception class.

    :param selector: A string of the format '0xffffffff' which indicates the
        4-byte ABI function selector of a rich revert error type, which is
        expected to be found as a class attribute on some class in
        `exceptions_module`:code:.
    :param exceptions_module: The Python module in which to look for a class
        with a `selector`:code: attribute matching the value of the
        `selector`:code: argument.
    """
    # noqa: D202 (No blank lines allowed after function docstring
    def _get_rich_revert_exception_classes():
        def _exception_name_is_class_with_selector(name: str):
            if not isclass(getattr(exceptions_module, name)):
                return False

            try:
                getattr(exceptions_module, name).selector
            except AttributeError:
                return False

            return True

        def _convert_class_name_to_class(name: str):
            return getattr(exceptions_module, name)

        return list(
            map(
                _convert_class_name_to_class,
                filter(
                    _exception_name_is_class_with_selector,
                    dir(exceptions_module),
                ),
            )
        )

    rich_reverts = _get_rich_revert_exception_classes()

    try:
        return next(
            filter(
                lambda rich_revert: rich_revert.selector == selector,
                rich_reverts,
            )
        )
    except StopIteration:
        raise NoExceptionForSelector(selector)



"""Generated wrapper for Forwarder Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for Forwarder below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        ForwarderValidator,
    )
except ImportError:

    class ForwarderValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass


class Tuple0x6ca34a6f(TypedDict):
    """Python representation of a tuple or struct.

    Solidity compiler output does not include the names of structs that appear
    in method definitions.  A tuple found in an ABI may have been written in
    Solidity as a literal, anonymous tuple, or it may have been written as a
    named `struct`:code:, but there is no way to tell from the compiler
    output.  This class represents a tuple that appeared in a method
    definition.  Its name is derived from a hash of that tuple's field names,
    and every method whose ABI refers to a tuple with that same list of field
    names will have a generated wrapper method that refers to this class.

    Any members of type `bytes`:code: should be encoded as UTF-8, which can be
    accomplished via `str.encode("utf_8")`:code:
    """

    makerAddress: str

    takerAddress: str

    feeRecipientAddress: str

    senderAddress: str

    makerAssetAmount: int

    takerAssetAmount: int

    makerFee: int

    takerFee: int

    expirationTimeSeconds: int

    salt: int

    makerAssetData: Union[bytes, str]

    takerAssetData: Union[bytes, str]

    makerFeeAssetData: Union[bytes, str]

    takerFeeAssetData: Union[bytes, str]


class ApproveMakerAssetProxyMethod(ContractMethod):
    """Various interfaces to the approveMakerAssetProxy method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, asset_data: Union[bytes, str]):
        """Validate the inputs to the approveMakerAssetProxy method."""
        self.validator.assert_valid(
            method_name="approveMakerAssetProxy",
            parameter_name="assetData",
            argument_value=asset_data,
        )
        return asset_data

    def call(
        self,
        asset_data: Union[bytes, str],
        tx_params: Optional[TxParams] = None,
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Approves the respective proxy for a given asset to transfer tokens on
        the Forwarder contract's behalf. This is necessary because an order fee
        denominated in the maker asset (i.e. a percentage fee) is sent by the
        Forwarder contract to the fee recipient. This method needs to be called
        before forwarding orders of a maker asset that hasn't previously been
        approved.

        :param assetData: Byte array encoded for the respective asset proxy.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (asset_data) = self.validate_and_normalize_inputs(asset_data)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_data).call(tx_params.as_dict())

    def send_transaction(
        self,
        asset_data: Union[bytes, str],
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Approves the respective proxy for a given asset to transfer tokens on
        the Forwarder contract's behalf. This is necessary because an order fee
        denominated in the maker asset (i.e. a percentage fee) is sent by the
        Forwarder contract to the fee recipient. This method needs to be called
        before forwarding orders of a maker asset that hasn't previously been
        approved.

        :param assetData: Byte array encoded for the respective asset proxy.
        :param tx_params: transaction parameters
        """
        (asset_data) = self.validate_and_normalize_inputs(asset_data)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_data).transact(tx_params.as_dict())

    def estimate_gas(
        self,
        asset_data: Union[bytes, str],
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (asset_data) = self.validate_and_normalize_inputs(asset_data)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_data).estimateGas(
            tx_params.as_dict()
        )


class MarketBuyOrdersWithEthMethod(ContractMethod):
    """Various interfaces to the marketBuyOrdersWithEth method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(
        self,
        orders: List[Tuple0x6ca34a6f],
        maker_asset_buy_amount: int,
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
    ):
        """Validate the inputs to the marketBuyOrdersWithEth method."""
        self.validator.assert_valid(
            method_name="marketBuyOrdersWithEth",
            parameter_name="orders",
            argument_value=orders,
        )
        self.validator.assert_valid(
            method_name="marketBuyOrdersWithEth",
            parameter_name="makerAssetBuyAmount",
            argument_value=maker_asset_buy_amount,
        )
        # safeguard against fractional inputs
        maker_asset_buy_amount = int(maker_asset_buy_amount)
        self.validator.assert_valid(
            method_name="marketBuyOrdersWithEth",
            parameter_name="signatures",
            argument_value=signatures,
        )
        self.validator.assert_valid(
            method_name="marketBuyOrdersWithEth",
            parameter_name="feePercentage",
            argument_value=fee_percentage,
        )
        # safeguard against fractional inputs
        fee_percentage = int(fee_percentage)
        self.validator.assert_valid(
            method_name="marketBuyOrdersWithEth",
            parameter_name="feeRecipient",
            argument_value=fee_recipient,
        )
        fee_recipient = self.validate_and_checksum_address(fee_recipient)
        return (
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        )

    def call(
        self,
        orders: List[Tuple0x6ca34a6f],
        maker_asset_buy_amount: int,
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[Tuple[int, int, int], Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Attempt to buy makerAssetBuyAmount of makerAsset by selling ETH
        provided with transaction. The Forwarder may *fill* more than
        makerAssetBuyAmount of the makerAsset so that it can pay takerFees
        where takerFeeAssetData == makerAssetData (i.e. percentage fees). Any
        ETH not spent will be refunded to sender.

        :param feePercentage: Percentage of WETH sold that will payed as fee to
            forwarding contract feeRecipient.
        :param feeRecipient: Address that will receive ETH when orders are
            filled.
        :param makerAssetBuyAmount: Desired amount of makerAsset to purchase.
        :param orders: Array of order specifications used containing desired
            makerAsset and WETH as takerAsset.
        :param signatures: Proofs that orders have been created by makers.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        ) = self.validate_and_normalize_inputs(
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        ).call(tx_params.as_dict())

    def send_transaction(
        self,
        orders: List[Tuple0x6ca34a6f],
        maker_asset_buy_amount: int,
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Attempt to buy makerAssetBuyAmount of makerAsset by selling ETH
        provided with transaction. The Forwarder may *fill* more than
        makerAssetBuyAmount of the makerAsset so that it can pay takerFees
        where takerFeeAssetData == makerAssetData (i.e. percentage fees). Any
        ETH not spent will be refunded to sender.

        :param feePercentage: Percentage of WETH sold that will payed as fee to
            forwarding contract feeRecipient.
        :param feeRecipient: Address that will receive ETH when orders are
            filled.
        :param makerAssetBuyAmount: Desired amount of makerAsset to purchase.
        :param orders: Array of order specifications used containing desired
            makerAsset and WETH as takerAsset.
        :param signatures: Proofs that orders have been created by makers.
        :param tx_params: transaction parameters
        """
        (
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        ) = self.validate_and_normalize_inputs(
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        ).transact(tx_params.as_dict())

    def estimate_gas(
        self,
        orders: List[Tuple0x6ca34a6f],
        maker_asset_buy_amount: int,
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        ) = self.validate_and_normalize_inputs(
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(
            orders,
            maker_asset_buy_amount,
            signatures,
            fee_percentage,
            fee_recipient,
        ).estimateGas(tx_params.as_dict())


class MarketSellOrdersWithEthMethod(ContractMethod):
    """Various interfaces to the marketSellOrdersWithEth method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(
        self,
        orders: List[Tuple0x6ca34a6f],
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
    ):
        """Validate the inputs to the marketSellOrdersWithEth method."""
        self.validator.assert_valid(
            method_name="marketSellOrdersWithEth",
            parameter_name="orders",
            argument_value=orders,
        )
        self.validator.assert_valid(
            method_name="marketSellOrdersWithEth",
            parameter_name="signatures",
            argument_value=signatures,
        )
        self.validator.assert_valid(
            method_name="marketSellOrdersWithEth",
            parameter_name="feePercentage",
            argument_value=fee_percentage,
        )
        # safeguard against fractional inputs
        fee_percentage = int(fee_percentage)
        self.validator.assert_valid(
            method_name="marketSellOrdersWithEth",
            parameter_name="feeRecipient",
            argument_value=fee_recipient,
        )
        fee_recipient = self.validate_and_checksum_address(fee_recipient)
        return (orders, signatures, fee_percentage, fee_recipient)

    def call(
        self,
        orders: List[Tuple0x6ca34a6f],
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[Tuple[int, int, int], Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Purchases as much of orders' makerAssets as possible by selling as much
        of the ETH value sent as possible, accounting for order and forwarder
        fees.

        :param feePercentage: Percentage of WETH sold that will payed as fee to
            forwarding contract feeRecipient.
        :param feeRecipient: Address that will receive ETH when orders are
            filled.
        :param orders: Array of order specifications used containing desired
            makerAsset and WETH as takerAsset.
        :param signatures: Proofs that orders have been created by makers.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (
            orders,
            signatures,
            fee_percentage,
            fee_recipient,
        ) = self.validate_and_normalize_inputs(
            orders, signatures, fee_percentage, fee_recipient
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(
            orders, signatures, fee_percentage, fee_recipient
        ).call(tx_params.as_dict())

    def send_transaction(
        self,
        orders: List[Tuple0x6ca34a6f],
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Purchases as much of orders' makerAssets as possible by selling as much
        of the ETH value sent as possible, accounting for order and forwarder
        fees.

        :param feePercentage: Percentage of WETH sold that will payed as fee to
            forwarding contract feeRecipient.
        :param feeRecipient: Address that will receive ETH when orders are
            filled.
        :param orders: Array of order specifications used containing desired
            makerAsset and WETH as takerAsset.
        :param signatures: Proofs that orders have been created by makers.
        :param tx_params: transaction parameters
        """
        (
            orders,
            signatures,
            fee_percentage,
            fee_recipient,
        ) = self.validate_and_normalize_inputs(
            orders, signatures, fee_percentage, fee_recipient
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(
            orders, signatures, fee_percentage, fee_recipient
        ).transact(tx_params.as_dict())

    def estimate_gas(
        self,
        orders: List[Tuple0x6ca34a6f],
        signatures: List[Union[bytes, str]],
        fee_percentage: int,
        fee_recipient: str,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (
            orders,
            signatures,
            fee_percentage,
            fee_recipient,
        ) = self.validate_and_normalize_inputs(
            orders, signatures, fee_percentage, fee_recipient
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(
            orders, signatures, fee_percentage, fee_recipient
        ).estimateGas(tx_params.as_dict())


class OwnerMethod(ContractMethod):
    """Various interfaces to the owner method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method().estimateGas(tx_params.as_dict())


class TransferOwnershipMethod(ContractMethod):
    """Various interfaces to the transferOwnership method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(self, new_owner: str):
        """Validate the inputs to the transferOwnership method."""
        self.validator.assert_valid(
            method_name="transferOwnership",
            parameter_name="newOwner",
            argument_value=new_owner,
        )
        new_owner = self.validate_and_checksum_address(new_owner)
        return new_owner

    def call(
        self, new_owner: str, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(new_owner).call(tx_params.as_dict())

    def send_transaction(
        self, new_owner: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(new_owner).transact(tx_params.as_dict())

    def estimate_gas(
        self, new_owner: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(new_owner).estimateGas(
            tx_params.as_dict()
        )


class WithdrawAssetMethod(ContractMethod):
    """Various interfaces to the withdrawAsset method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self.underlying_method = contract_function

    def validate_and_normalize_inputs(
        self, asset_data: Union[bytes, str], amount: int
    ):
        """Validate the inputs to the withdrawAsset method."""
        self.validator.assert_valid(
            method_name="withdrawAsset",
            parameter_name="assetData",
            argument_value=asset_data,
        )
        self.validator.assert_valid(
            method_name="withdrawAsset",
            parameter_name="amount",
            argument_value=amount,
        )
        # safeguard against fractional inputs
        amount = int(amount)
        return (asset_data, amount)

    def call(
        self,
        asset_data: Union[bytes, str],
        amount: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Withdraws assets from this contract. The contract formerly required a
        ZRX balance in order to function optimally, and this function allows
        the ZRX to be withdrawn by owner. It may also be used to withdraw
        assets that were accidentally sent to this contract.

        :param amount: Amount of ERC20 token to withdraw.
        :param assetData: Byte array encoded for the respective asset proxy.
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (asset_data, amount) = self.validate_and_normalize_inputs(
            asset_data, amount
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_data, amount).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self,
        asset_data: Union[bytes, str],
        amount: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Withdraws assets from this contract. The contract formerly required a
        ZRX balance in order to function optimally, and this function allows
        the ZRX to be withdrawn by owner. It may also be used to withdraw
        assets that were accidentally sent to this contract.

        :param amount: Amount of ERC20 token to withdraw.
        :param assetData: Byte array encoded for the respective asset proxy.
        :param tx_params: transaction parameters
        """
        (asset_data, amount) = self.validate_and_normalize_inputs(
            asset_data, amount
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_data, amount).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        asset_data: Union[bytes, str],
        amount: int,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (asset_data, amount) = self.validate_and_normalize_inputs(
            asset_data, amount
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self.underlying_method(asset_data, amount).estimateGas(
            tx_params.as_dict()
        )


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class Forwarder:
    """Wrapper class for Forwarder Solidity contract.

    All method parameters of type `bytes`:code: should be encoded as UTF-8,
    which can be accomplished via `str.encode("utf_8")`:code:.
    """

    approve_maker_asset_proxy: ApproveMakerAssetProxyMethod
    """Constructor-initialized instance of
    :class:`ApproveMakerAssetProxyMethod`.
    """

    market_buy_orders_with_eth: MarketBuyOrdersWithEthMethod
    """Constructor-initialized instance of
    :class:`MarketBuyOrdersWithEthMethod`.
    """

    market_sell_orders_with_eth: MarketSellOrdersWithEthMethod
    """Constructor-initialized instance of
    :class:`MarketSellOrdersWithEthMethod`.
    """

    owner: OwnerMethod
    """Constructor-initialized instance of
    :class:`OwnerMethod`.
    """

    transfer_ownership: TransferOwnershipMethod
    """Constructor-initialized instance of
    :class:`TransferOwnershipMethod`.
    """

    withdraw_asset: WithdrawAssetMethod
    """Constructor-initialized instance of
    :class:`WithdrawAssetMethod`.
    """

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: ForwarderValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = ForwarderValidator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        if web3 is None:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                        middleware["function"], layer=middleware["layer"]
                    )
            except ValueError as value_error:
                if value_error.args == (
                    "You can't add the same un-named instance twice",
                ):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(
            address=to_checksum_address(contract_address), abi=Forwarder.abi()
        ).functions

        self.approve_maker_asset_proxy = ApproveMakerAssetProxyMethod(
            web3_or_provider,
            contract_address,
            functions.approveMakerAssetProxy,
            validator,
        )

        self.market_buy_orders_with_eth = MarketBuyOrdersWithEthMethod(
            web3_or_provider,
            contract_address,
            functions.marketBuyOrdersWithEth,
            validator,
        )

        self.market_sell_orders_with_eth = MarketSellOrdersWithEthMethod(
            web3_or_provider,
            contract_address,
            functions.marketSellOrdersWithEth,
            validator,
        )

        self.owner = OwnerMethod(
            web3_or_provider, contract_address, functions.owner, validator
        )

        self.transfer_ownership = TransferOwnershipMethod(
            web3_or_provider,
            contract_address,
            functions.transferOwnership,
            validator,
        )

        self.withdraw_asset = WithdrawAssetMethod(
            web3_or_provider,
            contract_address,
            functions.withdrawAsset,
            validator,
        )

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"address","name":"_exchange","type":"address"},{"internalType":"bytes","name":"_wethAssetData","type":"bytes"}],"payable":false,"stateMutability":"nonpayable","type":"constructor"},{"payable":true,"stateMutability":"payable","type":"fallback"},{"constant":false,"inputs":[{"internalType":"bytes","name":"assetData","type":"bytes"}],"name":"approveMakerAssetProxy","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"components":[{"internalType":"address","name":"makerAddress","type":"address"},{"internalType":"address","name":"takerAddress","type":"address"},{"internalType":"address","name":"feeRecipientAddress","type":"address"},{"internalType":"address","name":"senderAddress","type":"address"},{"internalType":"uint256","name":"makerAssetAmount","type":"uint256"},{"internalType":"uint256","name":"takerAssetAmount","type":"uint256"},{"internalType":"uint256","name":"makerFee","type":"uint256"},{"internalType":"uint256","name":"takerFee","type":"uint256"},{"internalType":"uint256","name":"expirationTimeSeconds","type":"uint256"},{"internalType":"uint256","name":"salt","type":"uint256"},{"internalType":"bytes","name":"makerAssetData","type":"bytes"},{"internalType":"bytes","name":"takerAssetData","type":"bytes"},{"internalType":"bytes","name":"makerFeeAssetData","type":"bytes"},{"internalType":"bytes","name":"takerFeeAssetData","type":"bytes"}],"internalType":"struct LibOrder.Order[]","name":"orders","type":"tuple[]"},{"internalType":"uint256","name":"makerAssetBuyAmount","type":"uint256"},{"internalType":"bytes[]","name":"signatures","type":"bytes[]"},{"internalType":"uint256","name":"feePercentage","type":"uint256"},{"internalType":"address payable","name":"feeRecipient","type":"address"}],"name":"marketBuyOrdersWithEth","outputs":[{"internalType":"uint256","name":"wethSpentAmount","type":"uint256"},{"internalType":"uint256","name":"makerAssetAcquiredAmount","type":"uint256"},{"internalType":"uint256","name":"ethFeePaid","type":"uint256"}],"payable":true,"stateMutability":"payable","type":"function"},{"constant":false,"inputs":[{"components":[{"internalType":"address","name":"makerAddress","type":"address"},{"internalType":"address","name":"takerAddress","type":"address"},{"internalType":"address","name":"feeRecipientAddress","type":"address"},{"internalType":"address","name":"senderAddress","type":"address"},{"internalType":"uint256","name":"makerAssetAmount","type":"uint256"},{"internalType":"uint256","name":"takerAssetAmount","type":"uint256"},{"internalType":"uint256","name":"makerFee","type":"uint256"},{"internalType":"uint256","name":"takerFee","type":"uint256"},{"internalType":"uint256","name":"expirationTimeSeconds","type":"uint256"},{"internalType":"uint256","name":"salt","type":"uint256"},{"internalType":"bytes","name":"makerAssetData","type":"bytes"},{"internalType":"bytes","name":"takerAssetData","type":"bytes"},{"internalType":"bytes","name":"makerFeeAssetData","type":"bytes"},{"internalType":"bytes","name":"takerFeeAssetData","type":"bytes"}],"internalType":"struct LibOrder.Order[]","name":"orders","type":"tuple[]"},{"internalType":"bytes[]","name":"signatures","type":"bytes[]"},{"internalType":"uint256","name":"feePercentage","type":"uint256"},{"internalType":"address payable","name":"feeRecipient","type":"address"}],"name":"marketSellOrdersWithEth","outputs":[{"internalType":"uint256","name":"wethSpentAmount","type":"uint256"},{"internalType":"uint256","name":"makerAssetAcquiredAmount","type":"uint256"},{"internalType":"uint256","name":"ethFeePaid","type":"uint256"}],"payable":true,"stateMutability":"payable","type":"function"},{"constant":true,"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"internalType":"address","name":"newOwner","type":"address"}],"name":"transferOwnership","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"internalType":"bytes","name":"assetData","type":"bytes"},{"internalType":"uint256","name":"amount","type":"uint256"}],"name":"withdrawAsset","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"}]'  # noqa: E501 (line-too-long)
        )


# pylint: disable=too-many-lines
